"""
    QuaO Project device_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..enum.provider_tag import ProviderTag
from ..enum.sdk import Sdk
from ..model.device.aws_braket_device import AwsBraketDevice
from ..model.device.ibm_cloud_device import IbmCloudDevice
from ..model.device.ibm_quantum_device import IbmQuantumDevice
from ..model.device.quao_braket_device import QuaoBraketDevice
from ..model.device.quao_qiskit_device import QuaoQiskitDevice
from ..model.device.quao_tytan_device import QuaoTytanDevice
from ..model.device.quao_dwave_device import QuaoDwaveDevice
from ..model.device.dwave_system_device import DwaveSystemDevice
from ..model.device.dwave_hybrid_device import DwaveHybridDevice
from ..model.provider.provider import Provider
from ..model.provider.dwave_system_provider import system_devices, hybrid_devices


class DeviceFactory:
    @staticmethod
    def create_device(
        provider: Provider, device_specification: str, authentication: dict, sdk: Sdk
    ):
        """
        @param sdk:
        @param provider:
        @param device_specification:
        @param authentication:
        @return:
        """

        provider_type = ProviderTag.resolve(provider.get_provider_type().value)

        if ProviderTag.QUAO_QUANTUM_SIMULATOR.__eq__(provider_type):
            if Sdk.QISKIT.__eq__(sdk):
                return QuaoQiskitDevice(provider, device_specification)

            if Sdk.BRAKET.__eq__(sdk):
                return QuaoBraketDevice(provider, device_specification)

            if Sdk.TYTAN.__eq__(sdk):
                return QuaoTytanDevice(provider, device_specification)

            if Sdk.D_WAVE_OCEAN.__eq__(sdk):
                return QuaoDwaveDevice(provider, device_specification)

        if ProviderTag.IBM_QUANTUM.__eq__(provider_type):
            return IbmQuantumDevice(provider, device_specification)

        if ProviderTag.IBM_CLOUD.__eq__(provider_type):
            return IbmCloudDevice(provider, device_specification)

        if ProviderTag.AWS_BRAKET.__eq__(provider_type):
            return AwsBraketDevice(
                provider,
                device_specification,
                authentication.get("bucketName"),
                authentication.get("prefix"),
            )

        if ProviderTag.D_WAVE.__eq__(provider_type):
            if device_specification in system_devices:
                return DwaveSystemDevice(provider, device_specification)

            if device_specification in hybrid_devices:
                return DwaveHybridDevice(provider, device_specification)

        raise Exception("Unsupported device!")



"""
    QuaO Project post_processing_task.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from .async_task import AsyncTask
from ..component.callback.update_job_metadata import update_job_metadata
from ..config.logging_config import logger
from ..data.promise.post_processing_promise import PostProcessingPromise
from ..data.response.job_response import JobResponse
from ..enum.media_type import MediaType
from ..enum.status.job_status import JobStatus
from ..enum.status.status_code import StatusCode
from ..util.json_parser_utils import JsonParserUtils


class PostProcessingTask(AsyncTask):
    def __init__(self,
                 post_processing_fn,
                 promise: PostProcessingPromise):
        self.post_processing_fn = post_processing_fn
        self.promise = promise

    def do(self):
        """
           Execute post_processing and send to backend
        """

        job_response = JobResponse(
            authentication=self.promise.authentication,
            project_header=self.promise.project_header,
            job_status=JobStatus.DONE.value,
            status_code=StatusCode.DONE)

        update_job_metadata(job_response=job_response,
                            callback_url=self.promise.callback_url.on_start)

        job_response.content_type = MediaType.APPLICATION_JSON

        try:

            logger.info("Execute post_processing ...")
            job_result_post_processing = self.post_processing_fn(self.promise.job_result)

            logger.debug("Parsing job result....")
            job_response.job_result = JsonParserUtils.parse(job_result_post_processing)

            update_job_metadata(job_response=job_response,
                                callback_url=self.promise.callback_url.on_done)

        except Exception as exception:
            job_response.job_result = {
                "error": "Error when execute post_processing(): ",
                "exception": str(exception),
            }
            job_response.status_code = StatusCode.ERROR
            job_response.job_status = JobStatus.ERROR.value

            update_job_metadata(job_response=job_response,
                                callback_url=self.promise.callback_url.on_error)



"""
    QuaO Project ibm_quantum_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from qiskit import transpile

from ...data.device.circuit_running_option import CircuitRunningOption
from ...model.device.qiskit_device import QiskitDevice
from ...config.logging_config import logger


class IbmQuantumDevice(QiskitDevice):

    def _is_simulator(self) -> bool:
        logger.debug('[IBM Quantum] Get device type')

        return self.device.configuration().simulator

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug('[IBM Quantum] Create job with {0} shots'.format(options.shots))

        transpile_circuit = transpile(circuit, self.device)

        return self.device.run(transpile_circuit, shots=options.shots)



"""
    QuaO Project ibm_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit_ibm_runtime.accounts import ChannelType

from ...enum.provider_tag import ProviderTag
from ...model.provider.provider import Provider
from ...config.logging_config import *


class IbmCloudProvider(Provider):

    def __init__(self, api_key, crn):
        super().__init__(ProviderTag.IBM_CLOUD)
        self.api_key = api_key
        self.crn = crn
        self.channel = 'ibm_cloud'

    def get_backend(self, device_specification: str):
        logger.debug('[IBM Cloud] Get backend')

        provider = self.collect_provider()

        return provider.get_backend(name=device_specification)

    def collect_provider(self):
        logger.debug('[IBM Cloud] Connect to provider')

        return QiskitRuntimeService(channel= self.channel,
                                    token=self.api_key,
                                    instance=self.crn)



"""
    QuaO Project dwave_hybrid_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..device.dwave_device import DwaveDevice
from ...data.device.circuit_running_option import CircuitRunningOption
from ...config.logging_config import *


class DwaveHybridDevice(DwaveDevice):

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[Dwave System] Create job")

        return self.device.sample_bqm(circuit, time_limit=10)

    def _get_provider_job_id(self, job) -> str:
        logger.debug("[Dwave System] Get provider job id")

        return job.id

    def _get_job_result(self, job):
        logger.debug('[Dwave System] Get job result')

        return job.result()

    def _calculate_execution_time(self, job_result) -> float:
        logger.debug("[Dwave System] Calculate execution time")

        self.execution_time = (
            job_result.get('sampleset').info.get('run_time') / 1000
        )

        logger.debug(
            "[Dwave System] Execution time calculation was: {0} seconds".format(
                self.execution_time
            )
        )


"""
    QuaO Project logging_config.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import sys

from loguru import logger

logger.add(sink=sys.stderr, format="{level} : {time} : {message}: {process}",  level='DEBUG')




"""
    QuaO Project dwave_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import ABC

from ..device.device import Device
from ...data.device.circuit_running_option import CircuitRunningOption
from ...enum.status.job_status import JobStatus
from ...config.logging_config import *


class DwaveDevice(Device, ABC):

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[Dwave System] Create job")

        return self.device.sample(circuit)

    def _is_simulator(self) -> bool:
        logger.debug("[Dwave System] Get device type")

        return True

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.debug("[Dwave System] Produce histogram")

        return None

    def _get_provider_job_id(self, job) -> str:
        logger.debug("[Dwave System] Get provider job id")

        return job.info.get("problem_id")

    def _get_job_status(self, job) -> str:
        logger.debug("[Dwave System] Get job status")

        return JobStatus.DONE.value

    def _calculate_execution_time(self, job_result) -> float:
        logger.debug("[Dwave System] Calculate execution time")

        self.execution_time = (
            job_result.get("_info").get("timing").get("qpu_access_time") / 1000
        )

        logger.debug(
            "[Dwave System] Execution time calculation was: {0} seconds".format(
                self.execution_time
            )
        )

    def _get_job_result(self, job):
        logger.debug('[Dwave System] Get job result')

        return job


"""
    QuaO Project processing_unit.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.base_enum import BaseEnum


class ProcessingUnit(BaseEnum):
    CPU = 'CPU'
    GPU = 'GPU'
    QPU = 'QPU'



"""
    QuaO Project ibm_quantum_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_provider import IBMProvider

from ...enum.provider_tag import ProviderTag
from ...model.provider.provider import Provider
from ...config.logging_config import logger


class IbmQuantumProvider(Provider):
    def __init__(self, api_token):
        super().__init__(ProviderTag.IBM_QUANTUM)
        self.api_token = api_token

    def get_backend(self, device_specification: str):
        logger.debug('[IBM Quantum] Get backend')

        provider = self.collect_provider()

        return provider.get_backend(device_specification)

    def collect_provider(self):
        logger.debug('[IBM Quantum] Connect to provider')

        return IBMProvider(token=self.api_token)



"""
    QuaO Project invocation_handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..component.backend.invocation import Invocation
from ..data.request.invocation_request import InvocationRequest
from ..handler.handler import Handler


class InvocationHandler(Handler):
    def __init__(self, request_data: dict,
                 circuit_preparation_fn,
                 post_processing_fn):
        super().__init__(request_data, post_processing_fn)
        self.circuit_preparation_fn = circuit_preparation_fn

    def handle(self):
        invocation_request = InvocationRequest(self.request_data)

        backend = Invocation(invocation_request)

        backend.submit_job(circuit_preparation_fn=self.circuit_preparation_fn,
                           post_processing_fn=self.post_processing_fn)



"""
    QuaO Project aws_braket_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import boto3
from braket.aws import AwsSession, AwsDevice

from ...enum.provider_tag import ProviderTag
from ...model.provider.provider import Provider
from ...config.logging_config import *


class AwsBraketProvider(Provider):

    def __init__(self, aws_access_key, aws_secret_access_key, region_name):
        super().__init__(ProviderTag.AWS_BRAKET)
        self.aws_access_key = aws_access_key
        self.aws_secret_access_key = aws_secret_access_key
        self.region_name = region_name

    def get_backend(self, device_specification: str):
        logger.debug('[Aws Braket] Get backend')

        session = self.collect_provider()

        return AwsDevice(
            arn=device_specification,
            aws_session=session)

    def collect_provider(self):
        logger.debug('[Aws Braket] Connect to provider')

        session = boto3.Session(
            aws_access_key_id=self.aws_access_key,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.region_name)

        return AwsSession(boto_session=session)



"""
    QuaO Project request.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..response.authentication import Authentication
from ..response.project_header import ProjectHeader
from ...data.callback.callback_url import CallbackUrl


class Request:
    def __init__(self, request_data):
        self.provider_job_id = request_data.get("providerJobId")
        self.authentication: Authentication = Authentication(
            user_token=request_data.get("userToken"),
            user_identity=request_data.get("userIdentity"))
        self.execution = CallbackUrl(request_data.get("execution"))
        self.analysis: CallbackUrl = CallbackUrl(request_data.get("analysis"))
        finalization_callback = request_data.get("finalization")
        self.finalization: CallbackUrl = CallbackUrl(finalization_callback) \
            if finalization_callback \
            else finalization_callback
        project_header = request_data.get("projectHeader")
        self.project_header: ProjectHeader = ProjectHeader(name=project_header.get("name"),
                                                           value=project_header.get("value"))


"""
    QuaO Project response_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..data.response.job_response import JobResponse


class ResponseUtils:

    @staticmethod
    def generate_response(job_response: JobResponse) -> dict:
        if job_response:
            status_code = job_response.status_code.value
            body = {
                "providerJobId": job_response.provider_job_id,
                "jobStatus": job_response.job_status,
                "jobResult": job_response.job_result,
                "contentType": job_response.content_type.value,
                "histogram": job_response.job_histogram,
                "executionTime": job_response.execution_time
            }

        else:
            status_code = job_response.status_code.value
            body = "Error in function code. Please contact the developer."

        return {
            "statusCode": status_code,
            "body": body,
            "userIdentity": job_response.user_identity,
            "userToken": job_response.user_token,
            "projectId": job_response.project_header.value
        }



"""
    QuaO Project quao_braket_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.devices import LocalSimulator

from ...config.logging_config import logger
from ...model.provider.quao_provider import QuaoProvider


class QuaoBraketProvider(QuaoProvider):
    def get_backend(self, device_specification):
        logger.debug('[Quao Braket] Get backend')

        provider = self.collect_provider()

        device_names = provider.registered_backends()

        if device_names.__contains__(device_specification):
            return LocalSimulator(device_specification)

        raise Exception('Unsupported device')

    def collect_provider(self):
        logger.debug('[Quao Braket] Connect to provider')

        return LocalSimulator()



"""
    QuaO Project sdk.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..enum.base_enum import BaseEnum


class Sdk(BaseEnum):
    QISKIT = 'qiskit'
    BRAKET = 'braket'
    TYTAN = 'tytan'
    D_WAVE_OCEAN = "d-wave ocean"

    @staticmethod
    def resolve(sdk: str):
        for element in Sdk:
            if element.value.__eq__(sdk):
                return element

        raise Exception("Sdk type is not supported!")



"""
    QuaO Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import Options, Session, Sampler

from ...data.device.circuit_running_option import CircuitRunningOption
from ...model.device.qiskit_device import QiskitDevice
from ...config.logging_config import *


class IbmCloudDevice(QiskitDevice):
    def _is_simulator(self) -> bool:
        logger.debug('[IBM Cloud] Get device type')

        return self.device.configuration().simulator

    def _create_job(self, circuit,  options: CircuitRunningOption):
        logger.debug('[IBM Cloud] Create job with {0} shots'.format(options.shots))

        running_options = Options(optimization_level=1)
        running_options.execution.shots = options.shots

        with Session(service=self.provider.collect_provider(), backend=self.device) as session:
            sampler = Sampler(session=session, options=running_options)
            job = sampler.run(circuits=circuit)

            return job



"""
    QuaO Project file_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from zipfile import ZipFile, ZIP_DEFLATED
from io import BytesIO


class FileUtils:
    @staticmethod
    def zip(io_buffer_value, file_name):
        """

        @param io_buffer_value:
        @param file_name:
        @return:
        """

        zip_buffer = BytesIO()

        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr(file_name, io_buffer_value)

        return zip_buffer



"""
    QuaO Project braket_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ...config.logging_config import logger
from ...data.device.circuit_running_option import CircuitRunningOption
from ...enum.status.job_status import JobStatus
from ...model.device.quao_device import QuaoDevice


class QuaoBraketDevice(QuaoDevice):
    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug('[Quao Braket] Create job with {0} shots'.format(options.shots))

        import time

        start_time = time.time()

        job = self.device.run(task_specification=circuit, shots=options.shots)

        self.execution_time = time.time() - start_time

        return job

    def _produce_histogram_data(self, job_result) -> dict:
        logger.debug('[Quao Braket] Produce histogram')

        return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[Quao Braket] Get provider job id')

        return job.id

    def _get_job_status(self, job) -> str:
        logger.debug('[Quao Braket] Get job status')

        job_state = job.state()
        if JobStatus.COMPLETED.value.__eq__(job_state):
            job_state = JobStatus.DONE.value

        return job_state




"""
    QuaO Project invocation_request.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..callback.callback_url import CallbackUrl
from ..request.request import Request
from ...enum.processing_unit import ProcessingUnit


class InvocationRequest(Request):
    def __init__(self, request_data: dict):
        super().__init__(request_data)
        self.input = request_data.get("input")
        self.shots = request_data.get("shots")
        self.device_id = request_data.get("deviceId")
        self.device_selection_url = request_data.get("serverUrl")
        self.sdk = request_data.get("sdk").lower() if request_data.get("sdk") else None
        self.circuit_export_url = request_data.get("circuitExportUrl")
        self.processing_unit = (
            ProcessingUnit.GPU
            if ProcessingUnit.GPU.value.__eq__(request_data.get("processingUnit"))
            else ProcessingUnit.CPU
        )
        self.preparation = CallbackUrl(request_data.get("preparation"))



"""
    QuaO Project provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import abstractmethod, ABC

from ...enum.provider_tag import ProviderTag
from ...config.logging_config import logger


class Provider(ABC):

    def __init__(self, provider_type: ProviderTag):
        self.provider_type = provider_type

    @abstractmethod
    def get_backend(self, device_specification):
        """
        @param device_specification:
        """

        raise NotImplemented('[Provider] get_backend() method must be implemented')

    @abstractmethod
    def collect_provider(self):
        """

        """

        raise NotImplemented('[Provider] collect_provider() method must be implemented')

    def get_provider_type(self):
        logger.debug('[Provider] Get provider type')

        return self.provider_type



"""
    QuaO Project job_fetching_handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from .handler import Handler
from ..component.backend.job_fetching import JobFetching
from ..data.request.job_fetching_request import JobFetchingRequest


class JobFetchingHandler(Handler):
    def __init__(self,
                 request_data: dict,
                 post_processing_fn):
        super().__init__(request_data, post_processing_fn)

    def handle(self):
        request = JobFetchingRequest(self.request_data)

        job_fetching = JobFetching(request)

        fetching_result = job_fetching.fetch(post_processing_fn=self.post_processing_fn)

        return fetching_result




"""
    QuaO Project thread_config.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from concurrent.futures import ThreadPoolExecutor

circuit_exporting_pool = ThreadPoolExecutor(max_workers=10,
                                            thread_name_prefix='circuit-exporting-pool-')

circuit_running_pool = ThreadPoolExecutor(max_workers=20,
                                          thread_name_prefix='circuit-running-pool-')



"""
    QuaO Project update_job_metadata.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

import requests

from ...data.response.job_response import JobResponse
from ...util.http_utils import HttpUtils
from ...util.response_utils import ResponseUtils
from ...config.logging_config import logger


def update_job_metadata(
        job_response: JobResponse,
        callback_url: str):
    """

    @param job_response: Job response
    @param callback_url: Callback url to Quao server
    """
    logger.info("Calling to backend update job metadata: {0}".format(callback_url))

    request_body = ResponseUtils.generate_response(job_response)

    try:
        response = requests.patch(
            url=callback_url,
            json={"data": request_body},
            headers=HttpUtils.create_application_json_header(
                token=job_response.user_token,
                project_header=job_response.project_header),
        )

        logger.info("Calling to backend with status code: {0}".format(response.status_code))

    except Exception as exception:
        logger.error(
            "Error calling to backend with exception: {0}".format(str(exception))
        )



"""
    QuaO Project circuit_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from braket.circuits import Circuit
from qiskit import QuantumCircuit
from dimod.binary.binary_quadratic_model import BinaryQuadraticModel


class CircuitUtils:
    @staticmethod
    def get_depth(circuit):
        """

        @param circuit:
        @return:
        """
        if isinstance(circuit, Circuit):
            return circuit.depth

        if isinstance(circuit, QuantumCircuit):
            return circuit.depth()

        raise Exception("Invalid circuit type!")

    @staticmethod
    def get_qubit_amount(circuit):
        """

        @param circuit:
        @return:
        """
        if isinstance(circuit, Circuit):
            return circuit.qubit_count

        if isinstance(circuit, QuantumCircuit):
            return int(circuit.num_qubits)
        
        if isinstance(circuit, BinaryQuadraticModel): 
            return 100

        raise Exception("Invalid circuit type!")



"""
    QuaO Project media_type.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.base_enum import BaseEnum


class MediaType(BaseEnum):
    APPLICATION_JSON = 'application/json'
    ALL_TYPE = '*/*'
    MULTIPART_FORM_DATA = 'multipart/form-data'
    APPLICATION_ZIP = 'application/zip'
    SVG_XML = 'image/svg+xml'



"""
    QuaO Project circuit_running_option.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ...enum.processing_unit import ProcessingUnit


class CircuitRunningOption:
    def __init__(self,
                 shots: int = 1024,
                 processing_unit: ProcessingUnit = ProcessingUnit.CPU,
                 executor=None,
                 max_job_size=None):
        self.shots = shots
        self.processing_unit = processing_unit
        self.executor = executor
        self.max_job_size = max_job_size
